#include "includes.h"
#include "AgesEditor_GUI.h"

#pragma once

class BackedPanel : public wxPanel
{
public:
	wxBitmap mybitmap;
	virtual void OnPaint(wxPaintEvent &event);
	BackedPanel(wxWindow *parent, wxWindowID winid=wxID_ANY, 
		const wxPoint &pos = wxDefaultPosition, const wxSize &size = wxDefaultSize, 
		long style=524288|wxBORDER_NONE,const wxString &name=wxPanelNameStr)
		: wxPanel(parent,winid,pos,size,style,name)
	{
		
	}
	void setBitmap(const wxBitmap &newbitmap)
	{
		mybitmap=newbitmap;
		Refresh();
	}
	DECLARE_EVENT_TABLE();
};

//typedef unsigned long int u32;
struct Cursor
{
	int x,y;
	int w,h;
	int thickness;
	bool visible;
	u8 r,g,b,a;
	Cursor()
	{
		x=0;y=0;w=0;h=0;thickness=1;visible=true;r=0,g=0,b=0,a=255;
	}
	bool operator==(const Cursor &other) const
	{
		return x==other.x && y==other.y && w==other.w && h==other.h && visible==other.visible;
	}
	bool operator!=(const Cursor &other) const
	{
		return !(*this==other);
	}
};

struct Grid
{
	Cursor grid;
	Cursor highlight;
	vector<Cursor> cursors;
	int height,width;
	Grid() : height(0),width(0) { }
};

class BackedPanelWithGrid : public BackedPanel
{
public:
	Grid mygrid;
	bool highlightvisible;

	int last_mousex;
	int last_mousey;

	int last_gridx;
	int last_gridy;

	int get_last_grid_x() const
	{
		int gridx_limit=mygrid.width/mygrid.grid.w;
		if (last_gridx<0 || last_gridx>=gridx_limit)
		{
			return -1;
		}
		return last_gridx;
	}
	int get_last_grid_y() const
	{
		int gridy_limit=mygrid.height/mygrid.grid.h;
		if (last_gridy<0 || last_gridy>=gridy_limit)
		{
			return -1;
		}
		return last_gridy;
	}

	BackedPanelWithGrid(wxWindow *parent, wxWindowID winid=wxID_ANY, 
		const wxPoint &pos = wxDefaultPosition, const wxSize &size = wxDefaultSize, 
		long style=524288|wxBORDER_NONE,const wxString &name=wxPanelNameStr)
		: BackedPanel(parent,winid,pos,size,style,name)
	{
		highlightvisible=false;
		Cursor &g=mygrid.grid;
		Cursor &h=mygrid.highlight;
		g.r=128;g.g=128;g.b=128;g.a=255;
		g.visible=false;
		g.thickness=1;
		g.w=16;
		g.h=16;

		h.r=255;h.g=255;h.b=255;h.a=255;
		h.visible=false;
		h.thickness=1;
		h.w=1;
		h.h=1;
		h.x=-1;
		h.y=-1;

		last_mousex=-1;
		last_mousey=-1;

		last_gridx=-1;
		last_gridy=-1;
	}
	static void drawcursor(wxDC &dc, const Cursor &c, const Cursor &g);
	static void draw(wxDC &dc, const Grid &mygrid);
	virtual void OnPaint(wxPaintEvent &event);
	void MoveCursor(Cursor &oldcursor, const Cursor &newcursor);
	virtual void OnMouse(wxMouseEvent &event);
	void setgrid(int gx, int gy, int nw=-1, int nh=-1);
	void UpdateHighlight(int x,int y,bool offscreen=false);
	DECLARE_EVENT_TABLE();
};
