//#include "stdafx.h"

//#include <assert.h>
#include "MyApp.h"
#include "bitmap.h"

#include <string>
typedef std::basic_string<_TCHAR> _tstring;

unsigned char VP<u8>::dummy=0xFF;
unsigned char CVP<u8>::dummy=0xFF;

//typedef u8 *BP;
//typedef const u8 *CBP;

void MyApp::load_vram_packets(int base,int alt, int animation)
{
	if (base==vram_base_number && alt==vram_alt_number && animation==vram_animation_number) return;
	clear_vram();
	load_vram_packet_base(base);
	load_vram_packet_alt(alt);
	load_vram_animation(animation);
}

void MyApp::load_vram_packet_alt(int number)
{
	vram_alt_number=number;
	
	if (number==0) return;
	int in_page=4;
	int in_address=(whichgame==GAME_AGES)? 0x5B28 : 0x595E;
	int packet_address=getu16(in_page,in_address+number*2);
	int packet_page=in_page;
	load_vram_packet(packet_page,packet_address);
}
void MyApp::load_vram_packet_base(int number)
{
	vram_base_number=number;
	
	int in_page=1;
	int in_address=(whichgame==GAME_AGES) ? 0x69DA : 0x6926;
	int packet_address=getu16(in_page,in_address+number*2);
	int packet_page=in_page;
	load_vram_packet(packet_page,packet_address);
}
void MyApp::load_vram_animation(int number)
{
	vram_animation_number=number;
	if (number==255) return;
	int page=4;
	const int addr1= (whichgame==GAME_AGES)?0x5B52 : 0x59B0;
	const int addr2= (whichgame==GAME_AGES)?0x5BE9 : 0x5A48;

	CBP A=rom(page,addr1+number*2);
	int data1_addr=A.getu16();
	if (data1_addr<0x4000 || data1_addr>=0x8000) return;
	A=rom(page,data1_addr);
	
	int anim_speed[4];
	int anim_addr[4];
	int anim_number[4];


	int tile_animation_flags=*A++;
	int i;
	for (i=0;i<4;i++)
	{
		int a2_addr=A.getu16();
		if (a2_addr>=0x4000 && a2_addr<0x8000)
		{
			anim_speed[i]=*rom(page,a2_addr);
			anim_addr[i]=a2_addr+1;
		}
		else
		{
			if (tile_animation_flags & (1<<i))
			{
				anim_addr[i]=-1;
				anim_speed[i]=-1;
				tile_animation_flags&=~(1<<i);
			}
		}
	}
	
	//let's animate them once
	for (i=0;i<4;i++)
	{
		if (tile_animation_flags & (1 << i))
		{
			int addr=anim_addr[i];


			
			if (addr>=0x4000 && addr<0x8000)
			{
				CBP B=rom(page,addr);
				int nextspeed=0;
				int n;
				int frame_count=0;

				while (nextspeed!=255)
				{
					n=*B++;
					nextspeed=*B++;
					frame_count++;
				}
				int frame;
				B-=2;
				for (frame=frame_count-1;frame>=0;frame--)
				{
					n=*B++;
					nextspeed=*B++;
					B-=4;
					//if next speed == 255, it repeats backwards
					anim_number[i]=n;
					anim_speed[i]=nextspeed;

					CBP C=rom(page,addr2+n*6);
					int src_page=*C++;
					int src_addr=C.getu16be();
					int dest_addr=C.getu16be();
					int dest_page=dest_addr&1;
					dest_addr&=~0x0F;
					int length=*C++;

					if (dest_addr>=0x8000 && dest_addr<0xA000 && src_addr>=0x4000 && src_addr<0x8000)
					{
						BP dest=ram(dest_page,dest_addr);
						CBP src=rom(src_page,src_addr);
						int length16=(length+1)*16;
						while (length16>0)
						{
							if (dest.myindex==0x28C0)
							{
								int i=0;
							}
							*dest++=*src++;
							length16--;
						}
					}
				}
			}
			else
			{
				int i=0;
			}
		}
	}
}

void MyApp::load_vram_packet(int packet_page, int packet_address)
{
	int src_page, compression_method;
	int src_addr;
	int dest_addr;
	int dest_page;
	int length;
	bool chain=true;
	
	int packet_number=0;
	const int packet_number_limit=256; //artificial limit to prevent infinate loops

	if (packet_address<0x4000 || packet_address>=0x8000) return;
	
	CBP p=rom(packet_page,packet_address);

	for (packet_number=0;packet_number<packet_number_limit && chain;packet_number++)
	{
		int b;
		b=*p++;
		src_page=b&0x3F;
		compression_method=b>>6;
		src_addr=p.getu16be();
		dest_addr=p.getu16be();
		const int destpage_mask=0x0F;
		dest_page=dest_addr&destpage_mask;
		dest_addr&=~destpage_mask;
		b=*p++;
		length=b&0x7F;
		length++;
		chain=0!=(b&0x80);
		decompress_vram_packet(compression_method,src_page,src_addr,dest_page,dest_addr,length);
		
		
//		packet_address+=6;
	}
}
void MyApp::lzdecompress(BP dest, CBP src,int length, bool LWM)
{
	int bytesleft=1;
	u8 mask=0;
	while (length>0)
	{
		bytesleft--;
		mask<<=1;
		if (bytesleft==0)
		{
			bytesleft=8;
			mask=*src++;
		}
		if (mask & 0x80)
		{
			int lzdist1=0,lzdist2=0;
			int lzlength=0;
			if (LWM)
			{
				lzdist1=*src++;
				int a=*src++;
				lzdist2=a & 7;
				a=a>>3;
				if (a!=0)
				{
					lzlength=a+2;
				}
				else
				{
					lzlength=*src++;
					if (lzlength==0) lzlength=256;
				}
			}
			else
			{
				int a=*src++;
				lzdist1=a & 0x1F;
				lzdist2=0;
				a>>=5;
				if (a==0)
				{
					lzlength=*src++;
					if (lzlength==0) lzlength=256;
				}
				else
				{
					lzlength=a+1;
				}
			}
			int i;
			s16 lzdist16= (lzdist1^0xFF) | ((lzdist2^0xFF)<<8);
			BP src2=dest+ lzdist16;
			for (i=0;i<lzlength;i++)
			{
				*dest++=*src2++;
				length--;
				if (length==0) return;
			}
		}
		else
		{
			*dest++=*src++;
			length--;
		}
	}
	
}
void MyApp::unpack_16_common(BP*dest_p, CBP*src_p)
{
	BP &dest = *dest_p;
	CBP &src = *src_p;
	int length=16;
	u16 mask=src.getu16be();
	if (mask==0)
	{
		memcpy(&dest[0],&src[0],16);
		dest+=16;
		src+=16;
		return;
	}
	else
	{
		u8 common=*src++;
		int i;
		for (i=0;i<16;i++)
		{
			if (mask & 0x8000)
			{
				*dest++=common;
			}
			else
			{
				*dest++=*src++;
			}
			mask<<=1;
		}
	}
}



void MyApp::decompress_vram_packet(int compression_method,int src_page,int src_addr,int dest_page,int dest_addr,int length)
{
	if (!(src_addr>=0x4000 && src_addr<0x8000)) // if source is not in the ROM area
	{
		return;
	}
	dest_addr&=0x1FFF;
	dest_addr+=0x8000;
	if (!(dest_addr>=0x8000 && dest_addr<0xE000)) // if dest is not in the RAM,SRAM,VRAM area
	{
		return;
	}
	CBP src=rom(src_page,src_addr);
	BP dest=ram(dest_page,dest_addr);
	int length16=length*16;
	
	//confirm destination is in proper VRAM area
	switch (compression_method)
	{
	case 0: //RAW
		{
			memcpy(&dest[0],&src[0],length16);
		}
		break;
	case 1: //LZ
	case 3: //LZ with long word mode
		{
			bool LWM=compression_method==3;
			lzdecompress(dest,src,length16,LWM);
		}
		break;
	case 2: //16 bits and common bytes
		{
			while (length16>0)
			{
				unpack_16_common(&dest,&src);
				length16-=16;
			}
		}
		break;
	}
}

bool savefile(const u8* data, int size, const _TCHAR* filename)
{
	FILE *myfile;
	myfile=_tfopen(filename,_T("wb"));
	if (myfile)
	{
		int bytes_written;
		bytes_written=fwrite(data,1,size,myfile);
		fclose(myfile);
		if (bytes_written!=size)
		{
			return false;
		}
		return true;
	}
	else
	{
		return false;
	}
}

bool savefile(const vector<u8> &data,const _TCHAR* filename)
{
	return savefile(&data[0],data.size(),filename);
}

bool loadfile(vector<u8> &file,const _TCHAR* filename)
{
	FILE *myfile;
	myfile=_tfopen(filename,_T("rb"));
	if (myfile)
	{
		fseek(myfile,0,SEEK_END);
		long size=ftell(myfile);
		fseek(myfile,0,SEEK_SET);
		int bytes_read=0;
		file.resize(size);
		bytes_read=fread(&file[0],1,size,myfile);
		fclose(myfile);
		if (bytes_read!=size) return false;
	}
	else
	{
		return false;
	}
	return true;
}




void MyApp::load_tileset(int tileset)
{
	if (tileset_number==tileset) return;
	tileset_number=tileset;
	bool forcewalls=0!=(tileset&0x100);
	tileset&=0xFF;
	load_tileset_part1(tileset);
	load_tileset_part2();
	if (forcewalls)
	{
		load_tileset_force_walls();
	}
}

void MyApp::load_tileset_force_walls()
{
	//This made me WTF.  There is a specific check in the game to see if we are in the past overworld, NOT at the maku tree, then force tiles 64-127 to use color #0 instead of color #6.
	int tilesleft=64;
	int page=0x03;
	int addr=0xD204;
	
	while (tilesleft)
	{
		int i;
		for (i=0;i<4;i++)
		{
			int a=getramu8(page,addr+i);
			if ((a&0x07)==6)
			{
				getramu8(page,addr+i)=(a&0xF8);
			}
		}
		addr+=8;
		--tilesleft;
	}
}


void MyApp::load_tileset_part1(int tileset)
{
	u8 z;
	const int tableA_page=1;
	const int tableA_addr=(whichgame==GAME_AGES) ? 0x787E : 0x7964;
	const int tableB_page=1;
	const int tableB_addr=(whichgame==GAME_AGES) ? 0x7870 : 0x794E;

	//A = basic information block
	//	index for "B"
	//  rom bank for "C"
	//  address of "C" (BE)
	//  destination address (BE) + page number as least significant nibble
	//  Length and Chain flag
	//B = base address
	//  base page number and long offset flag
	//  base address (BE)
	//C = entry
	//  Literal bits (for each zero starting with least sig bit, copy 8 literals)
	//  Offset (if short offset, use most sig 4 bits as length)
	//  Length (long offset only)
	//  repeats until all bytes are extracted
	CBP A,B,C;
	BP D;
	A=rom(tableA_page,tableA_addr+tileset*2);
	A=rom(tableA_page,A.getu16());

	bool chain=true;
	int chain_entry_number=0;
	int CHAIN_LIMIT=256;	//prevent infinate loops with an artificial limit, most chains are no longer than 2
	while (chain && chain_entry_number<CHAIN_LIMIT)
	{
		chain_entry_number++;
		int B_index=*A++;
		int C_page=*A++;
		C_page&=0x3F;
		int C_addr=A.getu16be();
		int dest_addr=A.getu16be();
		int dest_page=dest_addr&0x0F;
		dest_addr&=0xFFF0;
		int length=A.getu16be();
		chain=(length&0x8000)!=0;
		length&=0x7FFF;
		
		B=rom(tableB_page,tableB_addr+B_index*2);
		B=rom(tableB_page,B.getu16());

		int src_page=*B++;
		int src_base=B.getu16be();
		bool long_offset_flag=(src_page&0x80)!=0;
		src_page&=0x3F;

		C=rom(C_page,C_addr);
		D=ram(dest_page,dest_addr);

		int bytes_remaining=length;

		if (bytes_remaining+dest_addr>=0xE000)
		{
			chain=false;
			break;
		}

		int bits_remaining=1;
		int mask=0;
		while (bytes_remaining>0)
		{
			bits_remaining--;
			mask>>=1;
			if (bits_remaining<=0)
			{
				bits_remaining=8;
				mask=*C++;
			}
			if ((mask&1)==0)
			{
//				int i;
//				for (i=0;i<8;i++)
//				{
					z=*D++=*C++;
					bytes_remaining--;
					if (bytes_remaining==0) goto escape;
//				}
//				bits_remaining=0;
			}
			else
			{
				int size=0;
				int offset=0;
				if (long_offset_flag==false)
				{
					offset=C.getu16();
					size=offset>>12;
					offset&=0xFFF;
				}
				else
				{
					size=*C++;
					offset=C.getu16();
				}
				if (src_base+offset+size>0x8000)
				{
					chain=false;
					goto escape;
				}
				CBP S=rom(src_page,src_base+offset);
				while (size>0)
				{
					size--;
					z=*D++=*S++;
					bytes_remaining--;
					if (bytes_remaining==0) goto escape;
				}
			}
		}
escape:
		int unused=0;
	}
}
void MyApp::load_tileset_part2()
{
	u8 z;
	//Part II: turn the bytes read into RAM into real tile data
	int src_page=3;
	int src_addr=0xDC00;
	int dest_page=3;
	int dest_addr=0xD000;
	int size=256;

	int base_page=(whichgame==GAME_AGES) ? 0x18 : 0x17;
	int base_addr=0x4004;

	int tilesource_addr=getu16(base_page,0x4000);
	int attribsource_addr=getu16(base_page,0x4002);
	
	CBP TileSource=rom(base_page,tilesource_addr);
	CBP AttribSource=rom(base_page,attribsource_addr);

	CBP S=ram(src_page,src_addr);
	BP D=ram(dest_page,dest_addr);
	while (size>0)
	{
		int N=S.getu16();
		CBP T=rom(base_page,base_addr);
		T+=N*3;
		int b1,b2,b3;
		b1=*T++;
		b2=*T++;
		b3=*T++;
		int tiledat=b1+(b2>>4)*256;
		int attribdat=b3+(b2&0x0F)*256;
		CBP TS,AS;
		TS = TileSource+tiledat*4;
		AS = AttribSource+attribdat*4;
		int i=0;
		for (i=0;i<4;i++)
		{
			z=*D++=*TS++;
		}
		for (i=0;i<4;i++)
		{
			z=*D++=*AS++;
		}
		size--;
	}
}

int MyApp::get_map_plane(int areaclass, int season) const
{
	CBP S=get_area_class_struct(areaclass,season);
	return S[6];
}

CBP MyApp::get_area_class_struct(int areaclass, int season) const
{
	int classes_page=0x04;
	int classes_addr=(whichgame==GAME_AGES)? 0x4F9C : 0x4C84;
	classes_addr+=areaclass*8;
	CBP S=rom(classes_page,classes_addr);
	if (whichgame==GAME_SEASONS)
	{
		if (*S==0xFF)
		{
			S++;
			S=rom(classes_page,S.getu16());
			S+=season*8;
		}
	}
	return S;
}

bool MyApp::area_class_has_seasons(int areaclass) const
{
	if (whichgame==GAME_AGES) return false;
	int classes_page=0x04;
	int classes_addr=(whichgame==GAME_AGES)? 0x4F9C : 0x4C84;
	classes_addr+=areaclass*8;
	CBP S=rom(classes_page,classes_addr);
	if (whichgame==GAME_SEASONS)
	{
		if (*S==0xFF)
		{
			return true;
		}
	}
	return false;


}


void MyApp::load_area_class(int areaclass, int season)
{
	int ac_s=areaclass+season*256;
	if (area_class_number==ac_s) return;
	area_class_number=ac_s;
	areaclass&=0x7F;
	
	CBP S=get_area_class_struct(areaclass,season);
	flags1_number=*S++;
	flags2_number=*S++;
	int alt_vram=*S++;
	int base_vram=*S++;
	int pal=*S++;
	int maptiles=*S++;
	int mapplane=*S++;
	int animation=*S++;
	
	load_palette(pal);
	load_vram_packets(base_vram,alt_vram,animation);
	
	if (ShouldForceWallColorChange(flags1_number,flags2_number,whichgame)) maptiles+=0x100;
	load_tileset(maptiles);
}
void MyApp::load_palette(int pal, bool force)
{
	if (!force && pal==palette_number) return;
	if (!force)
	{
		clear_palette();
		//these are hardcoded here because the game only loads new palettes, it leaves the old stuff alone
		load_palette(14,true);
		load_palette(15,true);
		palette_number=pal;
	}
	
	u8 z; //for debugging
	
	int palpoint_page=1;
	int palpoint_addr=(whichgame==GAME_AGES) ? 0x632C : 0x6290;
	
	int pal_addr= rom(palpoint_page,palpoint_addr+2*pal).getu16();
	if (pal_addr<0x4000 || pal_addr>=0x8000) return;
	CBP P=rom(palpoint_page, pal_addr);

	bool chain=true;
	const int MAX_CHAIN=255;  //just to prevent infinite loops
	int chain_number=0;
	
	
	while (chain && chain_number<MAX_CHAIN)
	{
		chain_number++;

		int A=*P++;
		chain=(A&0x80)!=0;
		int num_pals=(A&7)+1;
		int dest_addr=0xDE80+(A&0x78);
		int dest_page=2;
		
		int src_addr=P.getu16();
		int src_page=(whichgame==GAME_AGES) ? 0x17 : 0x16;

		if (src_addr>=0x4000 && src_addr+num_pals * 8 <0x8000)
		{	
			CBP S=rom(src_page,src_addr);
			BP D=ram(dest_page,dest_addr);

			int c,p;
			for (p=0;p<num_pals;p++)
			{
				for (c=0;c<8;c++)
				{
					z=*D++=*S++;
				}
			}
		}
	}
	memcpy(&palette[0],&getramu8(2,0xDE80),128);
}

void make_vram_tile(bitmap& b, const u8 *t, int x0=0, int y0=0)
{
	int y,x;
	for (y=0;y<8;y++)
	{
		int b1=*t++;
		int b2=*t++;
		for (x=7;x>=0;x--)
		{
			int c=(b1 & 1) + (b2 & 1)*2;
			b[y+y0][x+x0]=c;
			b1>>=1;
			b2>>=1;
		}
	}
}

void MyApp::render_tiles()
{
	const int VRAM_LAST_TILE=0x400;
	//const int VRAM_LAST_TILE=0x380;

	u32 new_th=get_tiles_hash();
	if (tiles_hash==new_th) return;
	tiles_hash=new_th;

	int i;
	for (i=0;i<VRAM_LAST_TILE;i++)
	{
		if (i==0x180 && VRAM_LAST_TILE==0x380) i+=0x80;
		const u8 *t=&_vram[i*16];
		make_vram_tile(tiles[i],t,0,0);
	}
}
void MyApp::render_maptiles()
{
	render_tiles();
	
	u64 new_mth=get_maptiles_hash();
	if (maptiles_hash==new_mth) return;
	maptiles_hash=new_mth;

	CBP T=ram(3,0xD000);
	for (int i=0;i<256;i++)
	{
		for (int p=0;p<4;p++)
		{
			int t=T[0];
			int a=T[4];

			if (t<0x80) t+=0x100;
			if (a&0x08) t+=0x200;
			int c;
			c=a&7;
			bitmap mytile=tiles[t];
			mytile.add(c*4);
			if (a&0x20) mytile.hflip();
			if (a&0x40) mytile.vflip();
			
			int x=(p&1)*8;
			int y=(p>>1)*8;
			
			maptiles[i].blit(mytile,x,y,255);

			T++;
		}
		T+=4;
	}
}

void MyApp::render_palette()
{
	if (rgbpalette_palette==palette_number) return;
	rgbpalette_palette=palette_number;
	rgbpalette.resize((palette.size()/2)*3);
	int i;
	for (i=0;i<32;i++)
	{
		int c=palette[i*2]+palette[i*2+1]*256;
		int r,g,b;
		r=c&0x1F;
		g=(c>>5)&0x1F;
		b=(c>>10)&0x1F;
		r*=8;
		g*=8;
		b*=8;
		if (r==248)r=255;
		if (g==248)g=255;
		if (b==248)b=255;

		rgbpalette[i*3+0]=r;
		rgbpalette[i*3+1]=g;
		rgbpalette[i*3+2]=b;
	}
}

int MyApp::get_plane_width(int plane_in) const
{
	int plane,floor;
	get_floor_and_plane(plane_in,floor,plane);
	if (floor==-1) return 16;
//	if (floor==-2) return 8;
	return 8;
}
int MyApp::get_plane_height(int plane_in) const
{
	int plane,floor;
	get_floor_and_plane(plane_in,floor,plane);
	if (floor==-1) return 16;
//	if (floor==-2) return 8;
	return 8;
}
int MyApp::get_map_width(int plane_in) const
{
	return is_dungeon_map(plane_in)?15:10;
}
int MyApp::get_map_height(int plane_in) const
{
	return is_dungeon_map(plane_in)?11:8;
}
bool MyApp::is_dungeon_map(int plane_in) const
{
	return (get_plane(plane_in)>=4);
}

int MyApp::get_floor(int plane_in) const
{
	int floor_out,plane_out;
	get_floor_and_plane(plane_in,floor_out,plane_out);
	return floor_out;
}
int MyApp::get_plane(int plane_in) const
{
	int floor_out,plane_out;
	get_floor_and_plane(plane_in,floor_out,plane_out);
	return plane_out;
}
void MyApp::get_floor_and_plane(int plane_in, int &floor_out, int &plane_out) const
{
	if (whichgame==GAME_AGES)
	{
		if (plane_in<4)
		{
			plane_out=plane_in;
			floor_out=-1;
		}
		else
		{
			floor_out=plane_in-4;
			
			if (floor_out>=0x0E)
			{
				plane_out=5;
			}
			else
			{
				plane_out=4;
			}
			if (floor_out>=26)
			{
				plane_out=4+(floor_out&1);
				floor_out=-2-(floor_out-26);
//				floor_out=-2;
			}			
		}
	}
	if (whichgame==GAME_SEASONS)
	{
		floor_out=-1;
		if (plane_in<4)
		{
			plane_out=0;
		}
		else if (plane_in<5)
		{
			plane_out=plane_in-4+1;
		}
		else
		{
			floor_out=plane_in-5;
			if (floor_out>=0x0E)
			{
				plane_out=5;
			}
			else
			{
				plane_out=4;
			}
			if (floor_out>=22)
			{
				plane_out=5;
				if (floor_out==22) plane_out=4;
				floor_out=-2-(floor_out-22);
			}
		}
	}
}

int MyApp::get_map_number(int column, int row, int floor) const
{
	column&=15;
	row&=15;
	if (floor==-1) return column+row*16;
	column&=7;
	row&=7;
	if (floor<=-2)
	{
		int coord=column+row*8;
		//HARDCODED-in-the-editor FLOORS for those not found in the game's dungeon maps
		if (whichgame==GAME_AGES)
		{
			if (floor==-2)
			{
				if (coord<=4)
				{
					return 0xD0+coord;
				}
			}
			if (floor==-3)
			{
				if (coord<0x10)
				{
					return coord;
				}
				if (coord>=0x10 && coord<0x14)
				{
					return (coord-0x10)+0xAB;
				}
				if (coord>=0x18 && coord<=0x1D)
				{
					return (coord-0x18)+0xF6;
				}
			}
		}
		if (whichgame==GAME_SEASONS)
		{
			if (floor==-2)
			{
				if (coord<0x10)
				{
					return coord+0xE0;
				}
			}
			if (floor==-3)
			{
				if (coord<0x18)
				{
					return coord;
				}
				if (coord>=0x18)
				{
					return coord-0x18+0xB0;
				}
			}
		}
		return 0;
	}


	int coord=column+row*8;
	
	int floors_page=1;
	int floors_addr=(whichgame==GAME_AGES) ? 0x4FCE : 0x4F41;
	
	CBP floors=rom(floors_page,floors_addr);
	floors+=floor*64;
	floors+=coord;
	int room=*floors;
	return room;
}

int MyApp::get_area_class(int mapnumber, int plane) const
{
	//4:52D4
	int area_class_page=4;
	int area_class_addr=(whichgame==GAME_AGES) ? 0x52D4 : 0x533C;
	
	CBP class_ptr=rom(area_class_page,rom(area_class_page,area_class_addr + plane*2).getu16());
	class_ptr+=mapnumber;

	int areaclass=*class_ptr;
	return areaclass;
}

int MyApp::get_area_class(int column, int row, int plane_in) const
{
	int plane,floor;
	get_floor_and_plane(plane_in,floor,plane);
	int mapnumber=get_map_number(column,row,floor);
	int areaclass=get_area_class(mapnumber,plane)&0x7F;
	return areaclass;
}

int MyApp::get_season(int plane_in, int season_in) const
{
	if (whichgame==GAME_AGES) return 0;
	int season=season_in;
	if (season==-1)
	{
		if (plane_in>=0 && plane_in<=3)
		{
			season=plane_in;
		}
		else
		{
			season=0;
		}
	}
	return season;
}



void MyApp::load_map(int column, int row, int plane_in, bool add2, int season_in, bool loadareaclass)
{
	if (!romloaded()) return;

	int season=get_season(plane_in,season_in);

	int plane,floor;
	get_floor_and_plane(plane_in,floor,plane);

	if (add2)
	{
		plane|=2;
	}
	int mapnumber=get_map_number(column,row,floor);
	int areaclass=get_area_class(mapnumber,plane)&0x7F;
	int mapplane=get_map_plane(areaclass,season);
	if (loadareaclass)
	{
		load_area_class(areaclass,season);
	}
	load_map(mapnumber,mapplane);
}

void MyApp::load_map(int mapnumber, int mapplane, bitmap &Map) const
{
	//4:4F6C
	int base_ptr_page=4;
	int base_ptr_addr=(whichgame==GAME_AGES) ? 0x4F6C : 0x4C4C;
	
	base_ptr_addr+=mapplane*8;
	CBP ptr=rom(base_ptr_page,base_ptr_addr);

	int smallmap=*ptr++;
	int ptr_page=*ptr++;
	int ptr_addr=ptr.getu16();
	int base_page=*ptr++;
	int base_addr=ptr.getu16();

	if (smallmap)
	{
		const int W=10;
		const int H=8;
		Map.resize(W,H);
		CBP ptable=rom(ptr_page,ptr_addr);
		ptable+=mapnumber*2;
		int map_addr=ptable.getu16();
		int mode=map_addr>>14;
		map_addr&=0x3FFF;
		map_addr+=base_addr;
		int map_page=base_page;
		while (map_addr>=0x8000)
		{
			map_page++;
			map_addr-=0x4000;
		}
		//map addr will probably be outside the range of 4000-7FFF, but that's OK for now...
		CBP src=rom(map_page,map_addr);
		if (mode==0) //uncompressed
		{
			int r=0,c=0;
			while (r<H)
			{
				Map[r][c]=*src++;
				c++;
				if (c>=W)
				{
					c=0;
					r++;
				}
			}
		}
		else if (mode==1) //8 common
		{
			int r=0,c=0;
			int bitsleft=1;
			int mask=0;
			int common=0;
			while (r<H)
			{
				bitsleft--;
				mask>>=1;
				if (bitsleft==0)
				{
					bitsleft=8;
					mask=*src++;
					if (mask!=0)
					{
						common=*src++;
					}
				}
				if (mask&1)
				{
					Map[r][c]=common;
				}
				else
				{
					Map[r][c]=*src++;
				}
				c++;
				if (c>=W)
				{
					c=0;
					r++;
				}
			}
		}
		else //16 common
		{
			int r=0,c=0;
			int bitsleft=1;
			int mask=0;
			int common=0;
			while (r<H)
			{
				bitsleft--;
				mask>>=1;
				if (bitsleft==0)
				{
					bitsleft=16;
					mask=src.getu16();
					if (mask!=0)
						common=*src++;
				}
				if (mask&1)
				{
					Map[r][c]=common;
				}
				else
				{
					Map[r][c]=*src++;
				}
				c++;
				if (c>=W)
				{
					c=0;
					r++;
				}
			}
		}
	}
	else
	{
		const int W=16;
		const int H=11;
		Map.resize(W,H);
		CBP ptable=rom(ptr_page,ptr_addr);
		CBP dictionary=ptable;
		ptable+=0x1000;
		ptable+=2*mapnumber;
		int map_addr=ptable.getu16();
		map_addr+=base_addr;
		map_addr-=0x200;
		int map_page=base_page;
		while (map_addr>=0x8000)
		{
			map_page++;
			map_addr-=0x4000;
		}

		CBP src=rom(map_page,map_addr);
		
		{
			int r=0,c=0;
			int bitsleft=1;
			int mask=0;
			while (r<H)
			{
				bitsleft--;
				mask>>=1;
				if (bitsleft==0)
				{
					bitsleft=8;
					mask=*src++;
				}
				if (mask&1)
				{
					int dict_src=src.getu16();
					int size=(dict_src>>12)+3;
					dict_src&=0x0FFF;
					CBP src2=dictionary+dict_src;
					while (size>0)
					{
						Map[r][c]=*src2++;
						c++;
						if (c>=W)
						{
							c=0;
							r++;
						}
						if (r>=H) break;
						size--;
					}
				}
				else
				{
					Map[r][c]=*src++;
					c++;
					if (c>=W)
					{
						c=0;
						r++;
					}
					//if (r>=H) break;  //it's implicit
				}
			}
		}
	}
}
void MyApp::load_map(int mapnumber, int mapplane)
{
	if (mapplane>=8) return;
	if (map_number==mapnumber && mapplane==map_plane_number) return;
	map_number=mapnumber;
	map_plane_number=mapplane;
	return load_map(mapnumber,mapplane,Map);
}

/*
void MyApp::render_tilecolors(vector<u32> &colors, int areaclass)
{
	load_area_class(areaclass,season);
	render_maptiles();
	render_palette();
	int t,x,y;
	for (t=0;t<256;t++)
	{
		const bitmap &M=maptiles[t];
		int rsum=0,gsum=0,bsum=0;
		for (y=0;y<16;y++)
		{
			for (x=0;x<16;x++)
			{
				int c=M[y][x];
				u8 *pe=&rgbpalette[c*3];
				rsum+=pe[0];
				gsum+=pe[1];
				bsum+=pe[2];
			}
		}
		int r,g,b;
		r=rsum/256;
		g=gsum/256;
		b=bsum/256;
		colors[t]=r+(g<<8)+(b<<16);
	}
}
*/

void MyApp::get_dimensions(int plane_in, int &world_w, int &world_h, int &map_w, int &map_h) const
{
	world_w=get_plane_width(plane_in);
	world_h=get_plane_height(plane_in);
	map_w=get_map_width(plane_in);
	map_h=get_map_height(plane_in);
}

void MyApp::accumulate_tile_color(int vram_tile_number, int *sums) const
{
	const u8*t=&_vram[16*vram_tile_number];
	int x;
	int y;
	int b1,b2;
	for (y=0;y<8;y++)
	{
		b1=*t++;
		b2=*t++;
		for (x=0;x<8;x++)
		{
			int c = (b1&1)+2*(b2&1);
			sums[c]++;
			b1>>=1;
			b2>>=1;
		}
	}
}

u32 MyApp::get_tile_color(int tile) const
{
	int sums[32];
	memset(sums,0,sizeof(sums));
	CBP T=rom(3,0xD000);
	T+=tile*8;
	for (int p=0;p<4;p++)
	{
		int t=T[0];
		int a=T[4];
		if (t<0x80) t+=0x100;
		if (a&0x08) t+=0x200;
		int c=a&7;
		accumulate_tile_color(t,&sums[c*4]);
		T++;
	}
	int r=0,g=0,b=0;
	for (int i=0;i<32;i++)
	{
		r+=rgbpalette[i*3+0]*sums[i];
		g+=rgbpalette[i*3+1]*sums[i];
		b+=rgbpalette[i*3+2]*sums[i];
	}
	r/=256;
	g/=256;
	b/=256;
	return (r)+(g<<8)+(b<<16);
}

int MyApp::get_num_planes() const
{
	if (whichgame==GAME_AGES)
	{
		return 32;
	}
	else
	{
		return 29;
	}
}


void MyApp::create_minimaps(vector<bitmap32> &out_vec)
{
	out_vec.resize(get_num_planes());

	vector <u32> sortme;
	sortme.reserve((4*16*16)+(26*8*8));

	for (int plane_in=0;plane_in<(int)out_vec.size();plane_in++)
	{
		bitmap32 &out=out_vec[plane_in];
		out.resize(160,128);
		out.clear(0x808080);
	}
	
	for (int plane_in=0;plane_in<(int)out_vec.size();plane_in++)
	{
		int world_w,world_h,map_w,map_h;
		get_dimensions(plane_in,world_w,world_h,map_w,map_h);
		for (int wy=0;wy<world_h;wy++)
		{
			for (int wx=0;wx<world_w;wx++)
			{
				int areaclass=get_area_class(wx,wy,plane_in)&0x7F;
				areaclass+=get_season(plane_in,-1)*0x80;
				sortme.push_back((areaclass<<23)|(plane_in<<16)|(wy<<8)|(wx<<0));
			}
		}
	}
	sort(sortme.begin(),sortme.end());
	
	vector<u32> tile_colors;
	tile_colors.resize(256);
	memset(&tile_colors[0],0xFF,tile_colors.size()*sizeof(u32));

	int i;
	int last_area_class=-1;
	for (i=0;i<(int)sortme.size();i++)
	{
		u32 val=sortme[i];
		int areaclass=(val>>23)&0x1FF;
		int season=areaclass/0x80;
		int plane_in=(val>>16)&0x7F;
		int wy=(val>>8)&0xFF;
		int wx=(val>>0)&0xFF;
		
		int world_w,world_h,map_w,map_h;
		get_dimensions(plane_in,world_w,world_h,map_w,map_h);
		
		bitmap32 &out=out_vec[plane_in];

		u64 areahash=get_area_hash();
		load_map(wx,wy,plane_in,false,-1);
		if (areahash!=get_area_hash())
		{
			vector_fill(tile_colors,0xFFFFFFFF);
			render_palette();
		}

/*
		load_area_class(areaclass&0x7F
		
		if (last_area_class!=areaclass)
		{
			int prev_val=vram_base_number+(vram_alt_number<<8)+(palette_number<<16)+(tileset_number<<24);
			load_area_class(areaclass&0x7F,season);
			int new_val=vram_base_number+(vram_alt_number<<8)+(palette_number<<16)+(tileset_number<<24);
			if (new_val!=prev_val)
			{
				memset(&tile_colors[0],0xFF,tile_colors.size()*sizeof(u32));
				render_palette();
			}
		}
		last_area_class=areaclass;
		
		load_map(wx,wy,plane_in,false,-1,false);
*/		
		int x0=wx*map_w;
		int y0=wy*map_h;
		for (int my=0;my<(int)Map.getHeight();my++)
		{
			for (int mx=0;mx<(int)Map.getWidth();mx++)
			{
				int t=Map[my][mx];
				u32 color=tile_colors[t];
				if (color & 0x80000000)
				{
					color=tile_colors[t]=get_tile_color(t);
				}
				out[my+y0][mx+x0]=color;
			}
		}
	}
/*
	
		#if !_DEBUG
		
	//	std::map<int,std::map<int,int> >class_tile_color;
		vector<vector<u32> > class_tile_color;
		class_tile_color.resize(128);

		int wx,wy;
		int mx,my;
		
		for (wy=0;wy<world_h;wy++)
		{
			for (wx=0;wx<world_w;wx++)
			{
				int areaclass=get_area_class(wx,wy,plane_in)&0x7F;
				load_map(wx,wy,plane_in,false,false);
				vector<u32> &tilecolors=class_tile_color[areaclass];
				if (tilecolors.size()==0)
				{
					tilecolors.resize(256);
					render_tilecolors(tilecolors,areaclass);
				}

				int x0=wx*map_w;
				int y0=wy*map_h;
				for (my=0;my<map_h;my++)
				{
					for (mx=0;mx<map_w;mx++)
					{
						out[my+y0][mx+x0]=tilecolors[Map[my][mx]];
					}
				}
			}
		}
		#endif
	}
	*/
}

const _TCHAR *hexn(unsigned int n, int digits)
{
	int i;

	static _TCHAR hexbuffer[9];
	_TCHAR hextable[]=_T("0123456789ABCDEF");
	hexbuffer[8]=0;
	for (i=7;i>=8-digits;--i)
	{
		hexbuffer[i]=hextable[n&15];
		n>>=4;
	}
	return hexbuffer+8-digits;
}
const _TCHAR *hex2(unsigned int n)
{
	return hexn(n,2);
}

bool MyApp::export_tileset(const _TCHAR *filename)
{
	render_maptiles();
	render_palette();
	bitmap bmp;
	drawtileset(bmp,maptiles);
	bool retval;
	retval=bmp.write_png(filename,rgbpalette);
	return retval;
}

bool MyApp::export_all_tilesets(const _TCHAR *basepath, int first, int limit)
{
	int areaclass;
	bool success=true;

	_tstring outbase=basepath;

	for (areaclass=first;areaclass<limit;areaclass++)
	{
		if (area_class_has_seasons(areaclass))
		{
			int season;
			for (season=0;season<4;season++)
			{
				_tstring outfile=outbase;
				outfile+=hex2(areaclass);
				outfile+=_T("_");
				outfile+=hexn(season+10,1);
				outfile+=_T(".png");

				load_area_class(areaclass,season);
				success=export_tileset(outfile.c_str());
				if (!success) return false;
			}
		}
		else
		{
			_tstring outfile=outbase;
			outfile+=hex2(areaclass);
			outfile+=_T(".png");
			
			load_area_class(areaclass);
			success=export_tileset(outfile.c_str());
			if (!success) return false;
		}
	}
	return true;
}
bool MyApp::export_map(const _TCHAR *filename)
{
	render_maptiles();
	render_palette();
	bitmap bmp;
	drawmap(bmp,maptiles,Map);
	return bmp.write_png(filename,rgbpalette);
}
bool MyApp::export_plane(const _TCHAR *filename, int plane_in)
{
	bitmap32 planebmp;
	bitmap32 bmp32;
	bitmap bmp8;

	int WW,WH,MW,MH;
	get_dimensions(plane_in,WW,WH,MW,MH);

	planebmp.resize(WW*MW*16,WH*MH*16);
	planebmp.clear(0x808080);
	
	vector<u32> sortme;
	sortme.reserve(WH*WW);
	int wy,wx;
	int ac;
	for (wy=0;wy<WH;wy++)
	{
		for (wx=0;wx<WW;wx++)
		{
			ac=get_area_class(wx,wy,plane_in);
			sortme.push_back((ac<<16)+(wy<<8)+(wx));

		}
	}
	std::sort(sortme.begin(),sortme.end());
	int i;
	for (i=0;i<(int)sortme.size();i++)
	{
		u32 M=sortme[i];
		wx=(M>>0)&0xFF;
		wy=(M>>8)&0xFF;
		load_map(wx,wy,plane_in);
		render_tiles();
		render_palette();
		render_maptiles();
		drawmap(bmp8,maptiles,Map);
		bmp8.make32bit(bmp32,&rgbpalette[0]);
		int y=wy*MH*16;
		int x=wx*MW*16;
		planebmp.blit(bmp32,x,y,0xFFFFFFFF);
	}
	return planebmp.write_png(filename);
}
bool MyApp::export_all_planes(const _TCHAR *basepath)
{
	int plane_in;
	bool success=true;

	_tstring outbase=basepath;

	for (plane_in=0;plane_in<get_num_planes();plane_in++)
	{
		_tstring outfile=outbase;
		outfile+=hex2(plane_in);
		outfile+=_T(".png");
		
		success=export_plane(outfile.c_str(),plane_in);
		if (!success) return false;
	}
	return true;

	return false;
}


void MyApp::drawtileset(bitmap &out, const vector<bitmap> &maptiles)
{
	out.resize(256,256);
	for (int r=0;r<16;r++)
	{
		for (int c=0;c<16;c++)
		{
			int x,y,t;
			x=c*16;
			y=r*16;
			t=r*16+c;
			out.blit(maptiles[t],x,y,255);
		}
	}
}

void MyApp::drawmap(bitmap &out,const vector<bitmap> &maptiles, const bitmap &map)
{
	int W=map.getWidth();
	int H=map.getHeight();
	if (W==16) W=15;

	out.resize(W*16,H*16);
	int r,c;
	for (r=0;r<H;r++)
	{
		for (c=0;c<W;c++)
		{
			int x,y,t;
			x=c*16;
			y=r*16;
			t=map[r][c];
			out.blit(maptiles[t],x,y,255);
		}
	}
}
