#pragma once

#include "includes.h"

#include <vector>
#include <algorithm>
using std::vector;

#include "BP.h"
#include "bitmap.h"

bool loadfile(vector<u8> &file,const _TCHAR* filename);

template <class t>
void vector_fill(vector<t> &fillthis,const t &value)
{
	size_t i;
	for (i=0;i<fillthis.size();i++)
	{
		fillthis[i]=value;
	}
}

class MyApp;

class MyAppParent
{
public:
	vector<u8> romfile;
	bool loadfile(const _TCHAR *filename)
	{
		bool retval;
		retval=::loadfile(romfile,filename);
		if (retval==false) return false;
		return true;
	}
};

class MyApp
{
public:
	enum Whichgame_t
	{
		GAME_UNKNOWN = -1,
		GAME_AGES = 1,
		GAME_SEASONS = 2
	};

public:
	const MyAppParent *parent;
	CBP _romfile;
	vector<u8> _vram;
	vector<u8> _sram;
	vector<u8> _ram;
	vector<u8> palette;
	bitmap Map;

	Whichgame_t whichgame;

	int area_class_number;
	int tileset_number;
	int vram_base_number;
	int vram_alt_number;
	int palette_number;
	int map_plane_number;
	int map_number;
	int flags1_number;
	int flags2_number;
	int vram_animation_number;

	vector<bitmap> tiles;
	u32 tiles_hash;
//	int tiles_vram1;
//	int tiles_vram2;
//	int tiles_vram3;
	vector<bitmap> maptiles;
	u32 maptiles_hash;
//	int maptiles_tileset;
//	int maptiles_vram1;
//	int maptiles_vram2;
//	int maptiles_vram3;
	vector<u8> rgbpalette;
	int rgbpalette_palette;

private:
	void constructor()
	{
		parent=NULL;
		whichgame=GAME_UNKNOWN;
		_vram.resize(16384);
		_ram.resize(16384);
		palette.resize(128);
		_sram.resize(8192);
		clear_vram();
		clear_ram();
		clear_palette();
		clear_sram();

		invalidate_state();

		init_tiles();
		init_maptiles();

		invalidate_graphics();
	}
public:
	MyApp()
	{
		constructor();
	}
	MyApp(MyAppParent const &newparent)
	{
		constructor();
		attach(newparent);
	}
	void attach(MyAppParent const &newparent)
	{
		invalidate_state();
		clear_vram();
		clear_ram();
		clear_palette();
		clear_sram();
		init_tiles();
		init_maptiles();
		invalidate_graphics();
		

		parent=&newparent;
		_romfile=parent->romfile;
		if (0==memcmp("ZELDA NAYRU",&_romfile[0x134],11))
		{
			whichgame=GAME_AGES;
		}
		if (0==memcmp("ZELDA DIN",&_romfile[0x134],10))
		{
			whichgame=GAME_SEASONS;
		}
	}
	bool romloaded() const
	{
		return _romfile.size()>0;
	}
	void invalidate_graphics()
	{
		tiles_hash=-1;
		maptiles_hash=-1;
		rgbpalette_palette=-1;
	}
	void invalidate_state()
	{
		area_class_number=-1;
		tileset_number=-1;
		vram_base_number=-1;
		vram_alt_number=-1;
		vram_animation_number=-1;
		palette_number=-1;
		map_plane_number=-1;
		map_number=-1;
		flags1_number=-1;
		flags2_number=-1;
	}

	void init_tiles()
	{
		tiles.resize(0x400);
		for (int i=0;i<(int)tiles.size();i++)
		{
			tiles[i].resize(8,8);
		}
		
	}
	void init_maptiles()
	{
		maptiles.resize(256);
		for (int i=0;i<(int)maptiles.size();i++)
		{
			maptiles[i].resize(16,16);
		}
	}

	static int getpageaddress(int page, int address)
	{
		return page*0x4000+address-0x4000;
	}
	static int getvrampageaddress(int page, int address)
	{
		return (page&1)*8192+address-0x8000;
	}
	static int getrampageaddress(int page, int address)
	{
		return (page&7)*4096+address-0xD000;
	}
	CBP rom(int page, int address) const
	{
		CBP retval(_romfile);
		if (address>=0 && address<0x4000)
		{
			retval=&_romfile[address];
			return retval;
		}
		if (address>=0x4000 && address<0x8000)
		{
			retval=&getu8(page,address);
			return retval;
		}
		if (address>=0x8000 && address<0xA000)
		{
			retval=_vram;
			retval=&getvramu8(page,address);
			return retval;
		}
		if (address>=0xA000 && address<0xC000)
		{
			retval=_sram;
			retval=&_sram[address&0x1FFF];
			return retval;
		}
		if (address>=0xC000 && address<0xD000)
		{
			retval=_ram;
			retval=&_ram[address&0x0FFF];
			return retval;
		}
		if (address>=0xD000 && address<0xE000)
		{
			retval=_ram;
			retval=&getramu8(page,address);
			return retval;
		}
		return retval;
	}
	BP ram(int page, int address)
	{
		BP retval;
		if (address>=0x8000 && address<0xA000)
		{
			retval=_vram;
			retval=&getvramu8(page,address);
			return retval;
		}
		if (address>=0xA000 && address<0xC000)
		{
			retval=_sram;
			retval=&_sram[address&0x1FFF];
			return retval;
		}
		if (address>=0xC000 && address<0xD000)
		{
			retval=_ram;
			retval=&_ram[address&0x0FFF];
			return retval;
		}
		if (address>=0xD000 && address<0xE000)
		{
			retval=_ram;
			retval=&getramu8(page,address);
			return retval;
		}
		return retval;
	}

	u8& getramu8(int page,int address)
	{
		int addr=getrampageaddress(page,address);
		assert (addr>=0 && addr<(int)_ram.size());
		return _ram[addr];
	}
	const u8& getramu8(int page,int address) const
	{
		int addr=getrampageaddress(page,address);
		assert (addr>=0 && addr<(int)_ram.size());
		return _ram[addr];
	}
	u8& getvramu8(int page,int address)
	{
		int addr=getvrampageaddress(page, address);
		assert (addr>=0 && addr<(int)_vram.size());
		return _vram[addr];
	}
	const u8& getvramu8(int page,int address) const
	{
		int addr=getvrampageaddress(page, address);
		assert (addr>=0 && addr<(int)_vram.size());
		return _vram[addr];
	}
	const u8& getu8(int page,int address) const
	{
		return getu8(getpageaddress(page,address));
	}
	const u8& getu8(int address) const
	{
		assert(address>=0 && address<_romfile.size());
		return _romfile[address];
	}
	u16 getu16(int page, int address)
	{
		return getu8(page,address)+getu8(page,address+1)*256;
	}
	u16 getu16(int address)
	{
		return getu8(address)+getu8(address+1)*256;
	}
	u16 getu16be(int page, int address)
	{
		return getu8(page,address)*256+getu8(page,address+1);
	}
	u16 getu16be(int address)
	{
		return getu8(address)*256+getu8(address+1);
	}

	void clear_vram()
	{
		vector_fill(_vram,(u8)0x00);
		vram_base_number=-1;
		vram_alt_number=-1;
	}
	void clear_maptiles_ram()
	{
		tileset_number=-1;
		memset(&getramu8(3,0xDC00),0x00,0x200);
		memset(&getramu8(3,0xD000),0x00,0x800);
	}
	void clear_ram()
	{
		tileset_number=-1;
		palette_number=-1;
		vector_fill(_ram,(u8)0x00);
	}
	void clear_sram()
	{
		vector_fill(_sram,(u8)0xFF);
	}
	void clear_palette()
	{
		palette_number=-1;
		vector_fill(palette,(u8)0xFF);
		memset(&getramu8(2,0xDE80),0xFF,0x80);
	}

	CBP get_area_class_struct(int areaclass, int season=0) const;

	int get_plane_width(int plane_in) const;
	int get_plane_height(int plane_in) const;
	int get_map_width(int plane_in) const;
	int get_map_height(int plane_in) const;
	bool is_dungeon_map(int plane_in) const;
	int get_season(int plane_in, int season_in) const;
	void MyApp::get_dimensions(int plane_in, int &world_w, int &world_h, int &map_w, int &map_h) const;
	int get_num_planes() const;

	int get_floor(int plane_in) const;
	int get_plane(int plane_in) const;
	void get_floor_and_plane(int plane_in, int &floor_out, int &plane_out) const;
	int get_map_number(int column, int row, int floor) const;	//gets a map number by reading from the floor table, or use multiplcation for floor -1
	void load_map(int mapnumber, int mapplane, bitmap &Map) const;
	void load_map(int mapnumber, int mapplane);
	void load_map(int column, int row, int plane_in, bool add2=false, int season_in=-1, bool loadareaclass=true);
	void load_area_class(int areaclass, int season=0);
	void load_palette(int pal, bool force=false);
	void load_tileset_part1(int tileset);
	void load_tileset_part2();
	void load_tileset_force_walls();
	static inline bool ShouldForceWallColorChange(int flags1, int flags2, Whichgame_t whichgame)
	{
		 return whichgame==GAME_AGES && (flags1&0x70)==0 && 0!=(flags2&0x80) && 0==(flags2&2);
	}

	void load_tileset(int number);
	void load_vram_packets(int base,int alt,int animation);
	void load_vram_packet_alt(int number);
	void load_vram_packet_base(int number);
	void load_vram_animation(int number);
	void load_vram_packet(int packet_page,int packet_address);
	void decompress_vram_packet(int compression_method,int src_page,int src_addr,int dest_page,int dest_addr,int length);
	static void MyApp::lzdecompress(BP dest, CBP src,int length, bool LWM);
	static void MyApp::unpack_16_common(BP*dest_p, CBP*src_p);

	void render_tiles();
	void render_maptiles();
	void render_palette();

	bool area_class_has_seasons(int areaclass) const;

	int get_area_class(int mapnumber, int plane) const;
	int get_area_class(int column, int row, int plane) const;
	//void render_tilecolors(vector<u32> &colors, int areaclass);
	void accumulate_tile_color(int vram_tile_number, int *sums) const;
	u32 get_tile_color(int t) const;

	void create_minimaps(vector<bitmap32> &out_vec);
//	void create_minimap(bitmap32 &out, int plane);

	int get_map_plane(int areaclass, int season=0) const;

	bool export_tileset(const _TCHAR *filename);
	bool export_all_tilesets(const _TCHAR *basepath, int first=0, int limit=128);
	bool export_map(const _TCHAR *filename);
	bool export_plane(const _TCHAR *filename, int plane_in);
	bool export_all_planes(const _TCHAR *basepath);

	u64 get_area_hash() const
	{
		return get_area_hash(vram_base_number,vram_alt_number,vram_animation_number,palette_number,tileset_number);
	}
	static u64 get_area_hash(u8 vram1, u8 vram2, u8 vram3, u8 pal, u16 tileset)
	{
		return ((u64)tileset<<32)+(pal<<24)+(vram3<<16)+(vram2<<8)+(vram1<<0);
	}
	u32 get_tiles_hash() const
	{
		return get_tiles_hash(vram_base_number,vram_alt_number,vram_animation_number);
	}
	static u32 get_tiles_hash(u8 vram1, u8 vram2, u8 vram3)
	{
		return (vram1<<0) + (vram2<<8) + (vram3<<16);
	}
	u64 get_maptiles_hash() const
	{
		return get_maptiles_hash(vram_base_number,vram_alt_number,vram_animation_number,tileset_number);
	}
	static u64 get_maptiles_hash(u8 vram1, u8 vram2, u8 vram3, u16 tileset)
	{
		return (vram1<<0) + (vram2<<8) + (vram3<<16) + ((u64)tileset<<32);
	}
	u64 get_map_hash() const
	{
		return (u64)get_area_hash()+(((u64)map_plane_number&0xFF)<<48) + (((u64)map_number&0xFF)<<56);
	}


	static void drawtileset(bitmap& out, const vector<bitmap> &maptiles);
	static void drawmap(bitmap &out, const vector<bitmap> &maptiles, const bitmap &map);

};

const _TCHAR *hexn(unsigned int n, int digits);
const _TCHAR *hex2(unsigned int n);
