#include "bitmap.h"
#include <png.h>

u8 _bitmap_row<u8>::garbage;

void copy_palette(vector<png_color> &png_palette, const vector<u8> &byte_palette);
void make_rowpointers(vector<const u8*> &rowpointers, const bitmap &bmp);
bool write_png_file(const bitmap &bmp, const vector<u8> &palette, const _TCHAR *filename);


void copy_palette(vector<png_color> &png_palette, const vector<u8> &byte_palette)
{
	png_palette.resize(byte_palette.size()/3);
	int i;
	for (i=0;i<png_palette.size();i++)
	{
		png_palette[i].red=byte_palette[i*3+0];
		png_palette[i].green=byte_palette[i*3+1];
		png_palette[i].blue=byte_palette[i*3+2];
	}
}

void make_rowpointers(vector<const u8*> &rowpointers, const bitmap &bmp)
{
	int i;
	rowpointers.resize(bmp.getHeight());
	for (i=0;i<bmp.getHeight();i++)
	{
		rowpointers[i]=&(bmp[i][0]);
	}
}

void copy_image(vector<const u8*>&rowpointers, bitmap &bmp, const bitmap32 &bmp32)
{
	const int W=bmp32.getWidth();
	const int H=bmp32.getHeight();
	rowpointers.resize(H);
	bmp.resize(W*3,H);
	int y,x;
	for (y=0;y<H;y++)
	{
		const u32* src=&bmp32[y][0];
		u8 *p=&bmp[y][0];
		rowpointers[y]=p;
		for (x=0;x<W;x++)
		{
			u32 c=*src++;
			*p++=((c>>0) & 0xFF);
			*p++=((c>>8) & 0xFF);
			*p++=((c>>16) & 0xFF);
		}
	}
}

bool write_png_file(const bitmap &bmp, const vector<u8> &palette, const _TCHAR *filename)
{
	vector<const u8*> rowpointers;
	vector<png_color> png_palette;

	bool no_error=false;
	
	FILE *file=NULL;
	png_structp png_ptr=NULL;
	png_infop info_ptr=NULL;
	png_bytep * row_pointers = NULL;
	
	file=_tfopen(filename,_T("wb"));
	if (!file) goto err;
	
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr) goto err;

	if (setjmp(png_jmpbuf(png_ptr))) goto err;
	
	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) goto err;
	
	png_init_io(png_ptr, file);
	
	if (setjmp(png_jmpbuf(png_ptr))) goto err;
	
	png_set_IHDR(png_ptr, info_ptr, bmp.getWidth(), bmp.getHeight(), 8, PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);


	copy_palette(png_palette,palette);

	png_set_PLTE(png_ptr, info_ptr, &png_palette[0], png_palette.size());

	png_write_info(png_ptr, info_ptr);
	
	make_rowpointers(rowpointers,bmp);
	png_write_image(png_ptr, (png_bytepp)&rowpointers[0]);
	
	png_write_end(png_ptr,info_ptr);
	
	no_error=true;
	
	err:
    png_destroy_write_struct(&png_ptr, &info_ptr);
	if (file)
	{
		fclose(file);
		file=NULL;
	}
	return no_error;
}

bool write_png_file(const bitmap32 &bmp, const _TCHAR *filename)
{
	vector<const u8*> rowpointers;
	bitmap bmp8;

	bool no_error=false;
	
	FILE *file=NULL;
	png_structp png_ptr=NULL;
	png_infop info_ptr=NULL;
	png_bytep * row_pointers = NULL;
	
	file=_tfopen(filename,_T("wb"));
	if (!file) goto err;
	
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr) goto err;

	if (setjmp(png_jmpbuf(png_ptr))) goto err;
	
	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) goto err;
	
	png_init_io(png_ptr, file);
	
	if (setjmp(png_jmpbuf(png_ptr))) goto err;
	
	png_set_IHDR(png_ptr, info_ptr, bmp.getWidth(), bmp.getHeight(), 8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
	png_write_info(png_ptr, info_ptr);

	copy_image(rowpointers,bmp8,bmp);
	png_write_image(png_ptr, (png_bytepp)&rowpointers[0]);
	
	png_write_end(png_ptr,NULL);
	
	no_error=true;
	
	err:
    png_destroy_write_struct(&png_ptr, &info_ptr);
	if (file)
	{
		fclose(file);
		file=NULL;
	}
	return no_error;
}
